#!/usr/bin/python
# Copyright (c) Members of the EGEE Collaboration. 2004.
# See http://www.eu-egee.org/partners/ for details on the copyright holders.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Alarm script for FTS webapp
# G. McCance 2006
# Checks that the FTS web-application inside tomcat
# is responding. The script will write an alarm file if
# the application is down for more than 2 attempts.

import os
import sys
import fnmatch
import re
import string
import getopt
import commands
import syslog
import time
import shutil
import threading
import time
import popen2
import signal
import socket

class runner( threading.Thread ):
   def __init__(self, command, timeout, debug):
      self.command = command
      self.timeout = timeout
      self.debug = debug
      threading.Thread.__init__ ( self )
   def run(self):
      if self.debug: log_message("Running: " + self.command)
      proc = popen2.Popen4(self.command)
      self.subProcessPid = proc.pid
      exitstatus = proc.wait()
      if os.WTERMSIG(exitstatus) == signal.SIGKILL: # it timed out
         return
      else:
         self.retcode = os.WEXITSTATUS(exitstatus)
         self.finished = 1
         return

   def runCommandWithTimeout(self):
      self.finished = 0
      self.start()
      tstart = time.time()
      while not self.finished:
         if (time.time() - tstart) > self.timeout: break
         time.sleep(0.1)
      if self.finished:
         if self.retcode == 0:
            return 0
         else:
            log_message("Got bad exit status from command")
            return 1
      else:
         if self.debug: log_message("Command timed out (killing it)")
         os.kill(self.subProcessPid, signal.SIGKILL)
         return 1


def usage():
   print 'Usage:'
   print '\t', sys.argv[0], '[ --debug ] [ vetofile=filename ] [ statefile=filename ] [ alarmfile=filename ]' 

def do_restart(agent):
   print "Restarting tomcat container daemon: " + agent

def log_message(message):
   message = time.strftime("%a, %d %b %Y %H:%M:%S +0000", time.gmtime()) + ": " + message
   print message   

def main():
	debug=False
 
        if os.geteuid() != 0:
	   log_message('You must run this script as root!')
	   sys.exit(1)

        if not os.path.isfile('/etc/tomcat5/Catalina/localhost/glite-data-transfer-fts.xml'):
	   log_message('No FTS web-application configured on this node')
	   sys.exit(0)

	try:
   	   opts = getopt.getopt(sys.argv[1:], '', [ 'debug', 'vetofile=', 'statefile=', 'alarmfile=' ])
	except getopt.GetoptError:
   	   usage()
   	   sys.exit(1)

	vetofilename='/etc/glite-fts-webservice.disabled'
	statefile='/var/lib/fts-stuck-alarm/fts-stuck.state'
	alarmfile='/var/lib/fts-stuck-alarm/fts-stuck.alarm'

        try:
           os.mkdir('/var/lib/fts-stuck-alarm')
        except OSError:
           pass
           # already exists

	for o,a in opts[0]:
	   if o == '--debug': debug=True
	   if o == '--vetofile':
	      vetofilename=a
	   if o == '--statefile':
	      statefile=a
	   if o == '--alarmfile':
	      alarmfile=a

	# Read veto list: agents in this list will be ignored by sensor
	if os.path.isfile(vetofilename):
	   log_message('FTS web-application is disabled on this node:\n' + vetofilename)
           sys.exit(0)

	# Snapshot process names
	procs=os.listdir('/proc')
	procs=fnmatch.filter(procs, '[0-9]*')
        tomcatfound = 0	
	for proc in procs:
	   fn='/proc/' + proc + '/cmdline'
	   try:
	      fil=open(fn, 'r')
	   except IOError:
	      continue
	   psname=fil.readline()
	   fil.close()
	   if psname is not None:
		psname = string.strip(psname)
		psname = string.translate(psname, string.maketrans(string.printable, string.printable), '\x00')
	        if psname.find('/usr/share/tomcat5/bin/bootstrap.jar') != -1:
		   tomcatfound = proc

	# Check lockfile
	if os.path.exists('/var/lock/subsys/fts-stuck-check'):
	   # read it and get pid
	   lockfile=open('/var/lock/subsys/fts-stuck-check', 'r')
	   lockpid=lockfile.readline()
	   lockpid=string.strip(lockpid)
	   lockfile.close()
	   # now check the pid existed a few microseconds ago
	   for i in procs:
	      if lockpid == i:
		 log_message('ERROR: lockfile exists and process exists.. not starting!')
		 sys.exit(5)
	   else:
		log_message('WARN: lockfile exists, but no process.. acquiring lock and running..')
	   os.remove('/var/lock/subsys/fts-stuck-check')

	lockfile=open('/var/lock/subsys/fts-stuck-check', 'w')
	lockfile.write(`os.getpid()` + '\n')
	lockfile.close()   

        if tomcatfound == 0:
           log_message('Tomcat container daemon is not running on this node. Alarm will not trigger.')
           sys.exit(1)

        if debug: log_message('Tomcat is running with PID: ' + tomcatfound)

        # do web-app test here
        webappbad = testwebapp(debug = debug)

        # always remove the alarm file
        if os.path.isfile(alarmfile):
	   os.unlink(alarmfile)

        # Clean the statefile and exit if it's running OK
        if not webappbad:
           if os.path.isfile(statefile):
	      os.unlink(statefile)
           if debug: log_message('Web application is available.')
           exitAndRemoveLock(0)

        # .. otherwise, increment the statefile and alarm if necessary
        num_attempts=0
        if os.path.isfile(statefile):
           statef = open(statefile,"r")
           webappstate = statef.readline()
           statef.close()
           num_attempts=string.atoi(webappstate)
           num_attempts = num_attempts + 1
           log_message('Web application is still down on attempt ' + `num_attempts`)

        # write alarm file if necessary
	if num_attempts > 2:
           log_message('Writing alarm file..')
           alarmf = open(alarmfile, 'w')
           alarmf.write('Problem with FTS wep-app running in tomcat container. PID: %s\n' % tomcatfound)	
           alarmf.close()
           num_attempts = 2

        # write the new state file
        statef = open(statefile,"w")
        statef.write(`num_attempts`)
        statef.close()


        exitAndRemoveLock(0)

def exitAndRemoveLock(code):
        # Delete lockfile
        os.remove('/var/lock/subsys/fts-stuck-check')
        sys.exit(code)

def testwebapp(debug = 1):
        endpoint = getEndpoint()
        command = "/opt/glite/bin/glite-transfer-channel-list -s %s > /dev/null 2>&1 < /dev/null " % endpoint
        runit = runner(command, timeout = 30, debug = debug)
        return runit.runCommandWithTimeout()

def getEndpoint():
        return "https://%s:8443/glite-data-transfer-fts/services/ChannelManagement" % socket.getfqdn()

if __name__ == "__main__":
    main()
