#! /usr/bin/python

#
# Copyright (c) Members of the EGEE Collaboration. 2004.
# See http://www.eu-egee.org/partners/ for details on the copyright holders.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
#  GLite Data Transfer Scripts
#
#  Authors: Paolo Badino <paolo.badino@cern.ch>
#  Version info: $Id: fts-report.py,v 1.13 2007/04/19 11:06:05 badino Exp $
#  Release: $Name:  $
#
#

import os,sys,getopt,string
import re
from time import gmtime, strftime, localtime, time
import cx_Oracle
from xml.dom.minidom import Document
from xml.dom.ext import PrettyPrint

#
# Report Builder Class
#
class ReportBuilder:
    #
    # Constructor
    #
    def __init__(self,r_type, i_type, ref_h, fts_i):

        self.connection    = None
        self.cursor        = None
        self.resultFile    = None
        self.reportType    = r_type
        self.intervalType  = i_type
        self.lowerIntervalBound = None
        self.upperIntervalBound = None
        self.nameFilter    = ''
        self.ftsInstance   = fts_i
        if(self.reportType != 'test' and self.intervalType != 'test'):
           self.initTimeInterval(ref_h)
 
    #
    # init TimeInterval
    # 
    def initTimeInterval(self, ref_h): 
        # common template for time interval bounds
        time_template      = "TO_TIMESTAMP_TZ((TO_CHAR(%s) || ' %s:00:00 ' || TO_CHAR(TZ_OFFSET(SESSIONTIMEZONE))) ,'YYYY-MM-DD HH24:MI:SS TZH:TZM')"
        time_template_hour = "TO_TIMESTAMP_TZ((TO_CHAR(%s) || ':00:00 ' || TO_CHAR(TZ_OFFSET(SESSIONTIMEZONE))) ,'YYYY-MM-DD HH24:MI:SS TZH:TZM')"
        if(self.intervalType == 'week'):
            # by default, take 8 a.m. as reference hour
            ref_hour  = '00'
            if(ref_h is not None):
                ref_hour = "%02d" % ref_h
            # consider all the transfers from last week 0am to today 0am
            self.lowerIntervalBound = time_template % ("SYSDATE  - INTERVAL '7' DAY,'YYYY-MM-DD'", ref_hour)
            self.upperIntervalBound = time_template % ("SYSDATE,'YYYY-MM-DD'", ref_hour)
            # Define filter for report filename
            self.nameFilter = '.... .. .._.. 00'
            self.typeDescription = 'Statistics concerning all the transfers performed during last week'
        elif(self.intervalType == 'day'):
            # by default, take 8 a.m. as reference hour
            ref_hour  = '08'
            if(ref_h is not None):
                ref_hour = "%02d" % ref_h
            # consider all the transfers from yesterday 0am to today 0am
            self.lowerIntervalBound = time_template % ("SYSDATE  - INTERVAL '1' DAY,'YYYY-MM-DD'", ref_hour)
            self.upperIntervalBound = time_template % ("SYSDATE,'YYYY-MM-DD'",ref_hour)
            # Define filter for report filename
            self.nameFilter = '.... .. .._.. 00'
            self.typeDescription = 'Statistics concerning all the transfers performed yesterday'
        elif(self.intervalType == 'hour'):
            # consider all the transfers from the hour previous the current one,
            # i.e. (current_hour -1):00 to current_hour:00)
            self.lowerIntervalBound = time_template_hour % ("SYSTIMESTAMP - INTERVAL '1' HOUR,'YYYY-MM-DD HH24'")
            self.upperIntervalBound = time_template_hour % ("SYSTIMESTAMP,'YYYY-MM-DD HH24'")
            # Define filter for report filename
            self.nameFilter = '.... .. .._.. 00'
            self.typeDescription = 'Statistics concerning all the transfers performed in last hour'
        elif(self.intervalType == 'currentday'):
            # consider all the transfers from today 0am to now
            self.lowerIntervalBound = time_template % ("SYSDATE,'YYYY-MM-DD'",'00')
            self.upperIntervalBound = None
            # Define filter for report filename
            self.nameFilter = '.... .. .._.. 00'
            self.typeDescription = 'Statistics concerning all the transfers performed today'
        elif(self.intervalType == 'currenthour'):
            # consider all the transfers in the current hour, 
            # i.e. (current_hour):00 to (current_hour):(current_minute)
            self.lowerIntervalBound = "SYSTIMESTAMP - INTERVAL '1' HOUR"
            self.upperIntervalBound = None
            # Define filter for report filename
            self.nameFilter = '.... .. .._.. ..'
            self.typeDescription = 'Statistics concerning all the transfers performed during the current hour'
        elif(self.intervalType == '15min'):
            # consider all the transfers in the last 15 minutes, 
            # i.e. (current_hour):(current_minutes - 15) to (current_hour):(current_minute)
            self.lowerIntervalBound = "SYSTIMESTAMP - INTERVAL '15' MINUTE"
            self.upperIntervalBound = None
            # Define filter for report filename
            self.nameFilter = '.... .. .._.. ..'
            self.typeDescription = 'Statistics concerning all the transfers performed in last 15 minutes'
        elif(self.intervalType == '30min'):
            # consider all the transfers in the last 30 minutes, 
            # i.e. (current_hour):(current_minutes - 30) to (current_hour):(current_minute)
            self.lowerIntervalBound = "SYSTIMESTAMP - INTERVAL '30' MINUTE"
            self.upperIntervalBound = None
            # Define filter for report filename
            self.nameFilter = '.... .. .._.. ..'
            self.typeDescription = 'Statistics concerning all the transfers performed in last 30 minutes'

        # Set the interval clause
        if((self.upperIntervalBound is not None) and (self.lowerIntervalBound is not None)):
            self.intervalClause = "SYS_EXTRACT_UTC(t_file.finish_time) < SYS_EXTRACT_UTC(%s) AND SYS_EXTRACT_UTC(t_file.finish_time) >= SYS_EXTRACT_UTC(%s)" % (self.upperIntervalBound, self.lowerIntervalBound)
        elif(self.lowerIntervalBound is not None):
            self.intervalClause = "SYS_EXTRACT_UTC(t_file.finish_time) >= SYS_EXTRACT_UTC(%s)" % (self.lowerIntervalBound)
        else:
            # Error
            print "Invalid interval type: %s" % (self.intervalType)
            sys.exit(-1)

    #
    # connect
    #
    def connect(self, u,p,c):
        try:
            self.connection = cx_Oracle.connect(u,p,c)
            self.cursor     = self.connection.cursor()
        except cx_Oracle.DatabaseError,detail:
            print "No Connection Established to Database \nERROR: %s" % (detail)
            sys.exit(-1)

    #
    # validate 
    #
    def validate(self):
        try:
            # Create a Query that check the DB schema is correct
            query = 'SELECT major, minor, patch FROM t_schema_vers WHERE ROWNUM <= 1' 
            self.cursor.execute(query);
            row = self.cursor.fetchone()
            major = int(row[0])
            minor = int(row[1])
            patch = int(row[2])
            if((major == 2 and minor >= 2) or (major == 3)):
                # Schema is ok
                print "Report Environment validation succeeded. Schema is %d.%d.%d" % (major,minor,patch)
            else:
                print "Unsupported schema. Current number is %d.%d.%d" % (major,minor,patch)
                sys.exit(-1) 
        except ValueError, detail:
            print "Invalid Schema Number\nERROR: %s" % (detail)
            sys.exit(-1)
        except cx_Oracle.DatabaseError,detail:
            print "No Connection Established to Database \nERROR: %s" % (detail)
            sys.exit(-1)
    #
    # generate 
    #
    def generate(self,output_dir):
        # Create Query
        query  = None
        partition_labels = list() 
        labels = list() 
        try:
            schema_query = 'SELECT major FROM t_schema_vers WHERE ROWNUM <= 1' 
            self.cursor.execute(schema_query);
            row = self.cursor.fetchone()
            major = int(row[0])
            if(major >= 3):
                # final state is Finished
                final_file_state = 'Finished'
            else:
                # final_state is Done
                final_file_state = 'Done'
        except ValueError, detail:
            print "Invalid Schema Number\nERROR: %s" % (detail)
            sys.exit(-1)
        except cx_Oracle.DatabaseError,detail:
            print "No Connection Established to Database \nERROR: %s" % (detail)
            sys.exit(-1)
        if(self.reportType == 'transfer'):
            query = '''
SELECT
    channel_name
  , vo_name
  , num_success
  , num_failure
  , DECODE(num_failure, 0, 0, ROUND( num_failure / (num_success + num_failure), 2) * 100) as failure_rate
  , first_failure_cause
  , DECODE(num_failure, 0, NULL,  ROUND (first_failure_count / num_failure, 2) * 100) as first_failure_rate
  , second_failure_cause
  , DECODE(num_failure, 0, NULL, ROUND (second_failure_count / num_failure, 2) * 100) as second_failure_rate
  , ROUND (avg_filesize / (1024 * 1024 * 1024), 2)                                    as avg_size_gb
  , ROUND (avg_duration , 2)                                                          as avg_dur
  , ROUND (avg_tx_rate / (1024 * 1024), 2)                                            as avg_tx_rate
  , ROUND (sum_effective_bytes / (1024 * 1024 * 1024), 2)                             as eff_bytes
  , ROUND (sum_bytes / (1024 * 1024 * 1024), 2)                                       as bytes
FROM (
  SELECT
      channel_name
    , vo_name
    , NVL(MAX(DECODE(transfer_state,'Completed',state_count, NULL)),0)          as num_success
    , NVL(MAX(DECODE(transfer_state,'Failed'   ,state_count, NULL)),0)          as num_failure
    ,     MAX(DECODE(top_reason,1,reason_class, NULL))                          as first_failure_cause
    ,     MAX(DECODE(top_reason,1,reason_count, NULL))                          as first_failure_count
    ,     MAX(DECODE(top_reason,2,reason_class, NULL))                          as second_failure_cause
    ,     MAX(DECODE(top_reason,2,reason_count, NULL))                          as second_failure_count
    , NVL(MAX(DECODE(transfer_state,'Completed', avg_filesize       , NULL)),0) as avg_filesize
    , NVL(MAX(DECODE(transfer_state,'Completed', avg_duration       , NULL)),0) as avg_duration
    , NVL(MAX(DECODE(transfer_state,'Completed', avg_tx_rate        , NULL)),0) as avg_tx_rate
    , NVL(MAX(DECODE(transfer_state,'Completed', sum_effective_bytes, NULL)),0) as sum_effective_bytes
    , NVL(MAX(DECODE(transfer_state,'Completed', sum_bytes          , NULL)),0) as sum_bytes
  FROM (
        SELECT
            channel_name
          , vo_name
          , transfer_state
          , reason_class
          , total_count
          , state_count
          , reason_count
          , DECODE(transfer_state,'Completed',99, DENSE_RANK() OVER (PARTITION BY channel_name, vo_name ORDER BY reason_count DESC, reason_class )) as top_reason
          , DECODE(state_count,0,0,sum_effective_bytes / state_count) as avg_filesize
          , DECODE(state_count,0,0,sum_vo_duration / state_count) as avg_duration
          , DECODE(state_count,0,0,sum_tx_rate / state_count) as avg_tx_rate
          , sum_effective_bytes
          , sum_bytes
        FROM (
                select
                   channel_name
                 , NVL(vo_name,'[All]') as vo_name
                 , transfer_state
                 , reason_class
                 , SUM(reason_count) OVER (PARTITION BY channel_name, vo_name) as total_count
                 , SUM(reason_count) OVER (PARTITION BY channel_name, vo_name, transfer_state) as state_count
                 , DECODE(transfer_state,'Completed',0 ,reason_count) as reason_count
                 , sum_tx_rate
                 , sum_vo_duration
                 , sum_effective_bytes
                 , sum_bytes
                FROM (
                        SELECT -- ROLLUP (vo_name)
                           channel_name
                         , vo_name
                         , transfer_state
                         , reason_class
                         , SUM(reason_vo_count) as reason_count
                         , SUM(sum_vo_duration) as sum_vo_duration
                         , SUM(sum_vo_tx_rate) as sum_tx_rate
                         , SUM(sum_vo_effective_bytes) as sum_effective_bytes
                         , SUM(sum_vo_bytes) as sum_bytes
                         FROM(
                                 SELECT
                                    channel_name
                                  , vo_name
                                  , transfer_state
                                  , reason_class
                                  , reason_vo_count
                                  , DECODE(transfer_state,'Completed',sum_vo_duration   ,NULL) as sum_vo_duration
                                  , DECODE(transfer_state,'Completed',sum_vo_tx_rate    ,NULL) as sum_vo_tx_rate
                                  , DECODE(transfer_state,'Completed',sum_vo_bytes_state,NULL) as sum_vo_effective_bytes
                                  , sum_vo_bytes
                                  FROM (
                                          SELECT DISTINCT
                                              t_job.channel_name  as channel_name
                                            , t_job.vo_name as vo_name
                                            , t_transfer.transfer_state as transfer_state
                                            , DECODE(t_transfer.transfer_state,'Completed','[Success]'
                                                                              ,t_transfer.reason_class) as reason_class
                                            , COUNT(*) OVER (PARTITION BY t_job.channel_name, t_job.vo_name , t_transfer.transfer_state, t_transfer.reason_class) as reason_vo_count
                                            , SUM(DECODE(t_transfer.transfer_state,'Completed',t_transfer.duration,0)) OVER (PARTITION BY t_job.channel_name, t_job.vo_name , t_transfer.transfer_state) as sum_vo_duration
                                            , SUM(DECODE(t_transfer.transfer_state,'Completed',t_file.filesize/t_transfer.duration,0)) OVER (PARTITION BY t_job.channel_name, t_job.vo_name , t_transfer.transfer_state) as sum_vo_tx_rate
                                            , SUM(DECODE(t_transfer.transfer_state,'Completed',t_file.filesize,t_transfer.bytes_written)) OVER (PARTITION BY t_job.channel_name, t_job.vo_name , t_transfer.transfer_state, t_transfer.reason_class) as sum_vo_bytes_state
                                            , SUM(DECODE(t_transfer.transfer_state,'Completed',t_file.filesize,t_transfer.bytes_written)) OVER (PARTITION BY t_job.channel_name, t_job.vo_name) as sum_vo_bytes
                                          FROM t_transfer,t_file,t_job
                                          WHERE t_transfer.file_id = t_file.file_id
                                          AND t_file.job_id = t_job.job_id
                                          AND t_file.file_state IN ('%FINAL_FILE_STATE%','Failed')
                                          AND t_transfer.transfer_state IN ('Completed','Failed')
                                          AND %INTERVAL_CLAUSE%
                                  --        AND SYS_EXTRACT_UTC(t_file.finish_time) >= SYS_EXTRACT_UTC(t_job.submit_time)
                                  --        AND SYS_EXTRACT_UTC(t_transfer.transfer_time) <= SYS_EXTRACT_UTC(t_file.finish_time)
                                  --        AND SYS_EXTRACT_UTC(t_transfer.transfer_time) >= SYS_EXTRACT_UTC(t_job.submit_time)
                                  )
                            ) group by channel_name, transfer_state, reason_class, ROLLUP (vo_name)
                        )
                )
        ) GROUP BY channel_name,vo_name
) ORDER BY channel_name, vo_name
'''
            # TODO Define namespace for the report        

            # Define Partitioning
            partition_labels.append(['channel'    , 'name' , "Channel Name"]) 
            partition_labels.append(['vo'         , 'name' , "VO Name"]) 

            # Set labels for the specific query
            labels.append(['num_success'          , "Number of successful transfers"])
            labels.append(['num_failure'          , "Number of failed transfers"])
            labels.append(['failure_rate'         , "Failure rate [%]"])
            labels.append(['first_failure_cause'  , "Most frequent failure reason"])
            labels.append(['first_failure_rate'   , "Rate of most frequent failure [%]"])
            labels.append(['second_failure_cause' , "Second frequent failure reason"])
            labels.append(['second_failure_rate'  , "Rate of second frequent failure [%]"])
            labels.append(['avg_filesize'         , "Average size of successful transfers [GB]"])
            labels.append(['duration'             , "Average duration of successful transfers [second]"])
            labels.append(['tx_rate'              , "Average Transfer rate for successful transfers [MB/second]"])
            labels.append(['sum_effective_bytes'  , "Number of effective bytes transferred [GB]"])
            labels.append(['sum_bytes'            , "Number of bytes transferred, including the one of failed transfers [GigaByte]"])

            # Set the Score label
            score_label = ['score'                , "The evaluated score for that channel"]
        elif(self.reportType == 'file'):
            query = '''
SELECT --by rollup
   channel_name
 , NVL(vo_name,'[All]') vo_name
 , SUM(first_attempt) first_attempt
 , SUM(second_attempt) second_attempt
 , SUM(third_attempt) third_attempt
 , SUM(further_attempts) further_attempts
 , SUM(failed_files) failed_files
FROM (
  SELECT
    channel_name
   , vo_name
   , MAX(DECODE(file_state,'%FINAL_FILE_STATE%',DECODE(num_failures,NULL,num_files,0),0)) first_attempt
   , MAX(DECODE(file_state,'%FINAL_FILE_STATE%',DECODE(num_failures,1,num_files,0),0)) second_attempt
   , MAX(DECODE(file_state,'%FINAL_FILE_STATE%',DECODE(num_failures,2,num_files,0),0)) third_attempt
   , SUM(DECODE(file_state,'%FINAL_FILE_STATE%',DECODE(num_failures,NULL,0,0,0,1,0,2,0,num_files),0)) further_attempts
   , MAX(DECODE(file_state,'Failed',num_files,0)) failed_files
  FROM (
    SELECT DISTINCT
      t_job.channel_name channel_name,
      t_job.vo_name vo_name,
      t_file.file_state file_state,
      t_file.num_failures num_failures,
      COUNT(t_file.file_state) OVER (PARTITION BY t_job.channel_name,t_job.vo_name,t_file.file_state,t_file.num_failures) num_files
    FROM t_file, t_job
    WHERE t_file.job_id = t_job.job_id
    AND t_file.file_state IN ('%FINAL_FILE_STATE%','Failed')
    AND t_job.channel_name IS NOT NULL
    AND %INTERVAL_CLAUSE%
  ) GROUP BY channel_name, vo_name
) GROUP BY channel_name, ROLLUP(vo_name)
'''
            # Define Partitioning
            partition_labels.append(['channel'    , 'name' , "Channel Name"]) 
            partition_labels.append(['vo'         , 'name' , "VO Name"]) 

            # Set labels for the specific query
            labels.append(['first_attempt'        , "Number of files successfully transfered at the first attempt"])
            labels.append(['second_attempt'       , "Number of files successfully transfered at the second attempt"])
            labels.append(['third_attempt'        , "Number of files successfully transfered at the third attempt"])
            labels.append(['further_attempts'     , "Number of files successfully transfered after the third attempt"])
            labels.append(['failed_files'         , "Number of failed files"])

            # Set the Score label
            score_label = ['score'                , "The evaluated score for that channel"]
        else:
            print "Unsupported Report Type: %s" % (self.reportType)
            sys.exit(-1)
      
        # Replace placeholders
        query = query.replace('%INTERVAL_CLAUSE%'   , self.intervalClause)
        query = query.replace('%FINAL_FILE_STATE%'  , final_file_state)
        #print "DEBUG: Executing query \n%s\n" % query 
        #return;

        # Process Query and generate Report
        self.process_query(output_dir, query, partition_labels, labels, score_label)

    #
    # Get the score for that row. The score is a float number in the interval [0(bad),1(good)]
    #
    def compute_score(self,row):
        score = 0.0
        if(self.reportType == 'transfer'):
            # For the moment, use a simple score: 1 - (failure_rate/100)
            score = 1 - (float(row[4])/100)
        elif(self.reportType == 'file'):
            # For the moment, use a simple score: 1 - (failure_rate/100)
            # where failure rate is failures/(success+failures)
            score = 1 - (float(row[6])/(row[2] + row[3] + row[4] + row[5] + row[6]))
        else:
            print "Unsupported Report Type: %s" % (self.reportType)
            sys.exit(-1)
        return score

    # 
    # process query and print result in the output file
    #
    def process_query(self, output_dir, query, partition_labels, labels, score_label):
        try:
            # Get the lower bound of the Time Interval
            time_query = "SELECT TO_CHAR(%s, 'YYYY-MM-DD HH24:MI:SS TZH:TZM') FROM DUAL" % self.lowerIntervalBound;
            self.cursor.execute(time_query);
            row = self.cursor.fetchone()
            lower = row[0];
            # Get the upper bound of the Time Interval
            if(self.upperIntervalBound is not None):
                time_query = "SELECT TO_CHAR(%s, 'YYYY-MM-DD HH24:MI:SS TZH:TZM') FROM DUAL" % self.upperIntervalBound;
            else:
                time_query = "SELECT TO_CHAR(SYSTIMESTAMP, 'YYYY-MM-DD HH24:MI:SS TZH:TZM') FROM DUAL"
            self.cursor.execute(time_query);
            row = self.cursor.fetchone()
            upper = row[0];

            # generate the report file name
            report_name = output_dir;
            if(report_name.endswith('/') == False):
                report_name = report_name + '/'
            source = str(upper);
            if(len(source) < len(self.nameFilter)):
                print "Cannot create report: invalid name filter [%s]" % (self.nameFilter)
                sys.exit(-1)
            # apply name filter        
            for i in range(0,len(self.nameFilter)):
                if self.nameFilter[i] == '.':
                    # take source
                    report_name = report_name + source[i]
                elif self.nameFilter[i] != ' ':
                    # take filter
                    report_name = report_name + self.nameFilter[i]
                #else:
                    # skip it
            report_name = report_name + ".xml"

            # Open output file
            try:
                self.resultFile = open(report_name, 'w')
            except IOError, detail:
                print "Cannot open file '%s' to write\nERROR: %s %s" % (report_name,IOError,detail)
                sys.exit(-1)

            # Execute Query
            self.cursor.execute(query)   
            rows = self.cursor.fetchall()

            # Print Report
            self.print_report(lower,upper,rows,partition_labels, labels, score_label)
            
        except cx_Oracle.DatabaseError,detail:
            print "Bad Query [%s]\nERROR: %s" % (query,detail)
            sys.exit(-1)

    #
    # print result of queries
    #
    def print_report(self,lower,upper,rows,partition_labels,labels, score_label):

        # Create the minidom document
        doc = Document()
        # Create the <report> base element
        root = doc.createElement("report")
        doc.appendChild(root)
        # Create the <details> element
        details = doc.createElement("details")
        root.appendChild(details)
        fts = doc.createElement("fts")
        fts.setAttribute("instance",self.ftsInstance)
        details.appendChild(fts)
        report_type = doc.createElement("type")
        report_type.appendChild(doc.createTextNode(self.reportType))
        details.appendChild(report_type)
        description = doc.createElement("description")
        description.appendChild(doc.createTextNode(self.typeDescription))
        details.appendChild(description)
        interval = doc.createElement("interval")
        interval.setAttribute("between",str(lower))
        interval.setAttribute("and",str(upper))
        details.appendChild(interval)
        # Create the <data> element
        data = doc.createElement("data")
        partition_count = len(partition_labels)
        root.appendChild(data)
        partition_nodes_list = list()
        for row in rows:
            # Create Partition nodes
            data_node = data
            i = 0
            for i in range(0, partition_count):
                # check if the partition tree exists
                if(i == len(partition_nodes_list)):
                    # stack is empty: create the partition tree
                    partition_node = doc.createElement(partition_labels[i][0])   
                    partition_node.setAttribute(partition_labels[i][1],row[i])
                    data_node.appendChild(partition_node)
                    # append the node to the list
                    partition_nodes_list.append(partition_node)
                    data_node = partition_node
                elif(i < len(partition_nodes_list)):
                    # Check the stack
                    p = partition_nodes_list[i]
                    # Check that it refers to the same node 
                    if(row[i] == p.getAttribute(partition_labels[i][1])):
                        partition_node = p
                        data_node = partition_node
                    else:
                        # Remove all the elements from the stack starting from i (included)
                        del partition_nodes_list[i:]
                        # Add the new Node
                        partition_node = doc.createElement(partition_labels[i][0])   
                        partition_node.setAttribute(partition_labels[i][1],row[i])
                        data_node.appendChild(partition_node)
                        # append the node to the list
                        partition_nodes_list.append(partition_node)
                        data_node = partition_node
                i = i + 1
            # Create Value nodes
            for i in range(0,len(labels)):
                data_value = doc.createElement(labels[i][0])
                value = row[partition_count + i]
                if(value is not None):
                    data_value.appendChild(doc.createTextNode(str(value)))
                data_node.appendChild(data_value)
               
            # Add a score node
            data_value = doc.createElement(score_label[0])
            data_value.appendChild(doc.createTextNode(str(self.compute_score(row))))
            data_node.appendChild(data_value)
             
        # Print our newly created XML
        PrettyPrint(doc, self.resultFile)

    #
    # disconnect
    #
    def disconnect(self):
        self.cursor.close()
        self.connection.close()
        if(self.resultFile is not None):
            self.resultFile.close()
        

###############################################################################

#
# Print usage
#
def help_message(str = ''):
    if str != '':
        print '\nERROR: %s' % (str)
        
    print '''
Syntax:
    %s -u <name> -p <password> -c <str> -t <report type> -i <interval type> [-v vo_name] [-o output directory] [-I FTS instance]
Options:
    -h     -- display this help
    -u     -- Database User's Name
    -p     -- Database User's Password
    -c     -- Database Connection string
    -t     -- The report type. It can be: 
                * transfer
                * file
    -i     -- The relevant interval. It can be:
                * week
                * day
                * hour
                * currentday
                * currenthour
                * 30min
                * 15min
    -r     -- The starting hour for the report, expressed as integer between 0 
              and 23 (Default is 8). This value is used only for the
              week and day interval
    -o     -- The name of the directory where the report file has to be generated 
    -I     -- The FTS instance name
''' % sys.argv[0]
    
    sys.exit(0)

#
# Main
#
def main():

    try:
        opts, args = getopt.getopt(sys.argv[1:],'hu:p:c:v:o:t:i:r:I:Z')
    except getopt.error, msg:
        help_message(msg)

    # default values
    user_name      = ''
    password       = ''
    connect_str    = ''
    interval_type  = ''
    report_type    = ''
    output_dir     = './'
    ref_hour       = ''    
    fts_instance   = ''    
    validation_mode= False
    # Get the arguments
    for o,a in opts[:]:
        if(o == '-u'):
            user_name = a
        elif(o == '-p'):
            password = a
        elif(o == '-c'):
            connect_str = a
        elif(o == '-t'):
            report_type = a
        elif(o == '-i'):
            interval_type = a
        elif(o == '-r'):
            ref_hour = a
        elif(o == '-o'):
            output_dir = a
        elif(o == '-I'):
            fts_instance = a
        elif(o == '-Z'):
            validation_mode = True
        elif(o == '-h'):
            help_message()

    # fake the parameters
    if(validation_mode == True):
        report_type   = 'test'
        interval_type = 'test'
      
    # Check the parameters
    if(len(user_name) == 0):
        help_message('No Username specified')
    if(len(password) == 0):
        help_message('No Password specified')
    if(len(connect_str) == 0):
        help_message('No DB Connection string specified')
    if(len(report_type) == 0):
        help_message('No report type specified')
    if(len(interval_type) == 0):
        help_message('No interval type specified')

    # check report type values
    if(report_type not in ('transfer','file','test')):
        help_message('Invalid report type specified')
    # check interval type values
    if(interval_type not in ('week','day','hour','currentday','currenthour','30min','15min','test')):
        help_message('Invalid interval type specified')
    # check reference hour value
    if(len(ref_hour) == 0):
        ref_hour = None
    else:
        # convert ref_hour to an integer
        try:
            ref_hour = int(ref_hour)
            if((ref_hour < 0) or (ref_hour > 23)):
                help_message('Invalid reference hour specified')
        except ValueError:
            help_message('Invalid reference hour specified: Not an integer')

    r = ReportBuilder(report_type,interval_type, ref_hour, fts_instance)
    # Create connection to the DB and create output file
    r.connect(user_name,password,connect_str)
    if(validation_mode == True):
        # validate
        r.validate()
    else: 
        # Generate Report
        r.generate(output_dir)
    
    # Close the connection to the DB and close output file
    r.disconnect()

    sys.exit(0)

if __name__ == "__main__":
    main()

